/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_CONV_H_
#define AIDGE_CORE_OPERATOR_CONV_H_

#include <array>
#include <cmath>    // std::floor
#include <cstddef>  // std::size_t
#include <string>
#include <utility>  // std::pair
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/ArrayHelpers.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Registrar.hpp" // SET_IMPL_MACRO
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class ConvAttr { StrideDims, DilationDims, KernelDims, NoBias };

template <DimIdx_t DIM>
class Conv_Op : public OperatorTensor,
                public Registrable<Conv_Op<DIM>, std::string, std::shared_ptr<OperatorImpl>(const Conv_Op<DIM> &)>,
                public StaticAttributes<ConvAttr,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        bool> {

public:
    static const std::string Type;

    Conv_Op() = delete;

    using Attributes_ = StaticAttributes<ConvAttr,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        bool>;
    template <ConvAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr Conv_Op(const std::array<DimSize_t, DIM> &kernelDims,
                      const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1),
                      bool noBias = false)
        : OperatorTensor(Type, 1, 2, 1),
          Attributes_(attr<ConvAttr::StrideDims>(strideDims),
                      attr<ConvAttr::DilationDims>(dilationDims),
                    //   attr<ConvAttr::InChannels>(inChannels),
                    //   attr<ConvAttr::OutChannels>(outChannels),
                      attr<ConvAttr::KernelDims>(kernelDims),
                      attr<ConvAttr::NoBias>(noBias)) {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Conv_Op(const Conv_Op<DIM>& op);

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Conv_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Conv_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? getInput(0) :
    //         (strcmp(inputName, "weight") ? getInput(1) :
    //         (strcmp(inputName, "bias") ? getInput(2) :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    // std::shared_ptr<Conv_Op> clone() const override final {

    // }

    bool forwardDims(bool /*allowDataDependency*/ = false) override final;

    std::vector<std::pair<std::vector<DimSize_t>, std::vector<DimSize_t>>>
    computeReceptiveField(const std::vector<DimSize_t>& firstEltDims,
                          const std::vector<DimSize_t>& outputDims,
                          const IOIndex_t outputIdx = 0) const override;


    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    DimSize_t inChannels() const {
        if (!getInput(1)) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Convolution operator has no weight Tensor associated so no specific number of input channel imposed.");
        }
        return getInput(1)->template dims<DIM+2>()[1];
    }

    DimSize_t outChannels() const {
        if (!getInput(1)) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Convolution operator has no weight Tensor associated so no specific number of input channel imposed.");
        }
        return getInput(1)->template dims<DIM+2>()[0];
    }

    static const std::vector<std::string> getInputsName(){
        return {"data_input", "weight", "bias"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};


/**
 * @brief Perform a convolution on the input Tensor.
 *
 * @tparam DIM Number of dimensions for the feature map.
 * @param inChannels Number of input channels.
 * @param outChannels Number of output channels.
 * @param kernelDims Dimensions of the kernel. Must be the same number of dimensions as the feature map.
 * @param name Name of the operator.
 * @param strideDims Dimensions of the stride attribute. Must be the same number of dimensions as the feature map.
 * @param dilationDims Dimensions of the dilation attribute. Must be the same number of dimensions as the feature map.
 * @return std::shared_ptr<Node> A Node containing the operator.
 */
template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Conv(DimSize_t inChannels,
                                  DimSize_t outChannels,
                                  const std::array<DimSize_t, DIM> &kernelDims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1),
                                  bool noBias = false) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    auto conv = std::make_shared<Node>(std::make_shared<Conv_Op<static_cast<DimIdx_t>(DIM)>>(kernelDims, strideDims, dilationDims, noBias), name);
    addProducer(conv, 1, append(outChannels, append(inChannels, kernelDims)), "w");
    addProducer(conv, 2, {(noBias ? 0 : outChannels)}, "b"); // already sets bias dims

    return conv;
}

// helper with C-style array instead of std::array for kernel_dims to allow automatic template DIM deduction
template <DimSize_t DIM>
inline std::shared_ptr<Node> Conv(
    DimSize_t inChannels,
    DimSize_t outChannels,
    DimSize_t const (&kernelDims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1),
    bool noBias = false) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    return Conv(inChannels, outChannels, to_array(kernelDims), name, strideDims, dilationDims, noBias);
}
}  // namespace Aidge

extern template class Aidge::Conv_Op<2>;

namespace {
template <>
const char *const EnumStrings<Aidge::ConvAttr>::data[] = {
    "StrideDims",
    "DilationDims",
    "KernelDims",
    "NoBias"
};
}

#endif /* AIDGE_CORE_OPERATOR_CONV_H_ */
