/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cstddef>  // std::size_t
#include <cstdint>  // std::int64_t
#include <string>
#include <vector>

#include "aidge/operator/Gather.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Types.h"

void Aidge::Gather_OpImpl::forward() {
    const Gather_Op& op = dynamic_cast<const Gather_Op&>(mOp);
    const auto axis = op.template getAttr<std::int8_t>("Axis");

    const std::size_t axisIdx = static_cast<std::size_t>(axis) + (axis >= 0 ? 0 : op.getInput(0)->dims().size());

    std::size_t postAxisElems = 1;
    for (std::size_t i = axisIdx + 1; i < op.getInput(0)->dims().size(); ++i) {
        postAxisElems *= op.getInput(0)->dims()[i];
    }
    std::size_t preAxisElems = 1;
    for (std::size_t i = 0; i < axisIdx; ++i) {
        preAxisElems *= op.getInput(0)->dims()[i];
    }

    std::size_t outputOffset = 0;
    for (std::size_t i=0; i<preAxisElems; ++i)
    {
        for(std::size_t j=0; j<op.template getAttr<std::vector<int64_t>>("Indices").size(); ++j)
        {
            const std::size_t idx = op.template getAttr<std::vector<int64_t>>("Indices")[j] >= 0 ?
                                        static_cast<std::size_t>(op.template getAttr<std::vector<int64_t>>("Indices")[j]) :
                                        static_cast<std::size_t>(op.template getAttr<std::vector<int64_t>>("Indices")[j] + static_cast<int>(op.getInput(0)->dims()[axisIdx]));
            op.getOutput(0)->getImpl()->copy(op.getInput(0)->getImpl()->rawPtr(i * postAxisElems * op.getInput(0)->dims()[axisIdx] + idx * postAxisElems), postAxisElems, outputOffset);
            outputOffset += postAxisElems;
        }
    }
}

const std::string Aidge::Gather_Op::Type = "Gather";

bool Aidge::Gather_Op::dimsForwarded() const {
    if (getInput(1) && !getInput(1)->empty()) {
        // output dims are data dependent
        return false;
    }

    return OperatorTensor::dimsForwarded();
}

bool Aidge::Gather_Op::forwardDims(bool allowDataDependency) {
    // check data input has been associated
    if (!getInput(0)) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "{}: input #0 should be associated with a Tensor", type());
    }

    if (getInput(0)->empty()) {
        return false;
    }

    if (getInput(1) && !getInput(1)->empty()) {
        if (!this->template getAttr<GatherAttr::Indices>().empty()) {
            Log::notice("Gather_Op: ignoring non-empty Indices attribute because input#1 takes precedence");
        }

        if (!allowDataDependency) {
            Log::warn("Gather_Op: unable to forwardDims() because output dims are data dependent on input#1");
            return false;
        }

        std::shared_ptr<Tensor> fallback;
        this->template getAttr<GatherAttr::GatheredShape>() = getInput(1)->dims();
        this->template getAttr<GatherAttr::Indices>().clear(); // If both are provided input would override attrs
        this->template getAttr<GatherAttr::Indices>().reserve(getInput(1)->size());
        const auto& indices = mInputs[1]->refCastFrom(fallback, NativeType<int64_t>::type, "cpu");
        std::copy_n(static_cast<int64_t*>(indices.getImpl()->hostPtr()),
                    indices.size(),
                    std::back_inserter(this->template getAttr<GatherAttr::Indices>()));
    }

    AIDGE_ASSERT(!this->template getAttr<GatherAttr::Indices>().empty(), "Missing input#1 or Indices attribute");

    std::vector<DimSize_t> outDims = getInput(0)->dims();

    std::int8_t axisIdx = this->template getAttr<GatherAttr::Axis>()>=0?
                            this->template getAttr<GatherAttr::Axis>():
                            this->template getAttr<GatherAttr::Axis>()+outDims.size();
    outDims.erase(outDims.begin() + static_cast<std::size_t>(axisIdx));
    if( !this->template getAttr<GatherAttr::GatheredShape>().empty())
    {
        outDims.insert(outDims.begin() + static_cast<std::size_t>(axisIdx),
                        this->template getAttr<GatherAttr::GatheredShape>().begin(),
                        this->template getAttr<GatherAttr::GatheredShape>().end());
    }
    mOutputs[0]->resize(outDims);
    return true;
}

void Aidge::Gather_Op::setBackend(const std::string& name, Aidge::DeviceIdx_t device) {
    if (Registrar<Gather_Op>::exists({name})) {
        SET_IMPL_MACRO(Gather_Op, *this, name);
    }
    else {
        mImpl = std::make_shared<Gather_OpImpl>(*this);
    }
    mOutputs[0]->setBackend(name, device);
}
